/*
 * Copyright (c) 2004 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: function.c,v 1.6 2003/11/22 11:37:29 zinnia Exp $
 */

#include "sps.h"

#include <ctype.h>

/* ---------------------------------------- */
/* utility functions */

static struct _coltbl {
	const char* name;
	Uint8 r;
	Uint8 g;
	Uint8 b;
} colortbl[] = {
	{"black", 0, 0, 0},
	{"white", 0xff, 0xff, 0xff},
	{"red", 0xff, 0, 0},
	{"green", 0, 0xff, 0},
	{"blue", 0, 0, 0xff},
	{"cyan", 0, 0xff, 0xff},
	{"yellow", 0xff, 0xff, 0},
	{"magenta", 0xff, 0, 0xff},
	{NULL, 0, 0, 0},
};


int gunistr_len(gunichar* src)
{
	gunichar* p;
	int result = 0;

	for(p = src; p != NULL && *p != 0; p++){
		result++;
	}
	return result;
}

gunichar* gunistr_copy(gunichar* dst, gunichar* src)
{
	gunichar* p;
	gunichar* q;
	for(p = src, q = dst; *p != 0; p++, q++){
		*q = *p;
	}
	*q = 0;
	return dst;
}

gunichar* gunistr_concat(gunichar* text1, gunichar* text2)
{
	gunichar* dst;
	dst = g_malloc(sizeof(gunichar) * (gunistr_len(text1) + gunistr_len(text2) + 1));
	gunistr_copy(dst, text1);
	gunistr_copy(dst + gunistr_len(text1), text2);
	return dst;
}

gunichar* gunistr_substring(gunichar* src, int offset, int len)
{
	int i;
	int len_remain;
	gunichar* dst;

	gunichar* q;

	dst = g_malloc(sizeof(gunichar) * (len + 1));

	q = dst;
	len_remain = len;
	for(i = 0; i < gunistr_len(src); i++){
		if(len_remain <= 0) break;
		if(i < offset) continue;

		*q = src[i];
		q++;
		len_remain--;
	}
	*q = 0;
	return dst;
}

static Uint8 GetNibble(char c)
{
	char p = tolower(c);
	if(p >= '0' && p <= '9'){
		return p - '0';
	} else if(p >= 'a' && p <= 'f') {
		return p - 'a' + 10;
	}
	return 0;
}

static Uint8 GetOctet(const char* expression, int offset)
{
	Uint8 result;
	result = GetNibble(*(expression + offset));
	result <<= 4;
	result |= GetNibble(*(expression + offset + 1)) & 0xf;

	return result;
}

Uint32 ParseColor(const char* text)
{
	int i; Uint8 r, g, b; int len;
	char tmp[7];

	for(i = 0; colortbl[i].name != NULL; i++){
		if(!strcasecmp(text, colortbl[i].name)) return SDL_MapRGB(context->surface->format, colortbl[i].r, colortbl[i].g, colortbl[i].b);

	}
	len = strlen(text);
	if(len > 6){
		strncpy(tmp, text, 6);
		tmp[6] = '\0';
	}
	else {
		int t;
		for(t = 0; t < 6; t++) tmp[t] = '0';
		strcpy(tmp + (6 - strlen(text)), text);
	}
	r = GetOctet(tmp, 0);
	g = GetOctet(tmp, 2);
	b = GetOctet(tmp, 4);

	fprintf(stderr, "%s -> %s -> %02x, %02x, %02x\n", text, tmp, r, g, b);
	
	return SDL_MapRGB(context->surface->format, r, g, b);
}

SPS_VERTICAL_ALIGN SPS_Vertical_Align_Parse(const char* expression)
{
	if(!strcasecmp(expression, "top")) return SPS_VERTICAL_ALIGN_TOP;
	if(!strcasecmp(expression, "bottom")) return SPS_VERTICAL_ALIGN_BOTTOM;
	if(!strcasecmp(expression, "center")) return SPS_VERTICAL_ALIGN_CENTER;

	return SPS_VERTICAL_ALIGN_BOTTOM;
}
		
SPS_HORIZONTAL_ALIGN SPS_Horizontal_Align_Parse(const char* expression)
{
	if(!strcasecmp(expression, "right")) return SPS_HORIZONTAL_ALIGN_RIGHT;
	if(!strcasecmp(expression, "left")) return SPS_HORIZONTAL_ALIGN_LEFT;
	if(!strcasecmp(expression, "center")) return SPS_HORIZONTAL_ALIGN_CENTER;

	return SPS_HORIZONTAL_ALIGN_LEFT;
}

static struct _ttfstyle {
	int character;
	int style;
} ttfstyle[] = {
	{'b', TTF_STYLE_BOLD},
	{'B', TTF_STYLE_BOLD},
	{'i', TTF_STYLE_ITALIC},
	{'I', TTF_STYLE_ITALIC},
	{'u', TTF_STYLE_UNDERLINE},
	{'U', TTF_STYLE_UNDERLINE},
	{0, 0},
};

int TTF_Style_Parse(const char* expression)
{
	int result = 0;
	int i;
	for(i = 0; ttfstyle[i].character != 0; i++)
	{
		if(strchr(expression, ttfstyle[i].character)){
			result |= ttfstyle[i].style;
		}
	}
	return result;
}
int RatioToPixelH(double h)
{
	int result = context->surface->w * h / 100;
	return result;
}

int RatioToPixelV(double v)
{
	int result = context->surface->h * v / 100;
	return result;
}

int RatioToPixel(double l)
{
	if(context->surface->w > context->surface->h){
		return RatioToPixelV(l);
	} else {
		return RatioToPixelH(l);
	}
}

char* SPS_State_ToString(SPS_STATE state)
{
	return state == SPS_STATE_PAGEWAIT ? "[pagewait]" :
		state == SPS_STATE_PAUSE ? "[pause]" :
		state == SPS_STATE_DONE ? "[done]" : 
		state == SPS_STATE_PAGELOAD ? "[pageload]" :
		"[unknown]" ;
}

void SPS_SetState(SPS_STATE newstate)
{
	SDL_mutexP(global.mutex);
	fprintf(stderr, "state: %s -> %s\n", SPS_State_ToString(global.state),
		SPS_State_ToString(newstate));
	global.state = newstate;
	SDL_mutexV(global.mutex);
	global.redraw = 1;
}

/* -------------------- */

static TTF_Font* GetFont(const char* fontindex)
{
	return g_hash_table_lookup(context->font, fontindex);
}
static SDL_Surface* GetIcon(const char* iconindex)
{
	return g_hash_table_lookup(context->icons, iconindex);
}
static void SDL_UnMapRGB(SDL_PixelFormat* fmt, SDL_Color* dst, Uint32 color)
{
	Uint8 r, g, b;
	SDL_GetRGB(color, fmt, &r, &g, &b);
	dst->r = r;
	dst->g = g;
	dst->b = b;
}


void showfps()
{
	char buf[0x100];
	SDL_Surface* surface;
	SDL_Color c;
	SDL_Rect r;
	TTF_Font* font;
	Uint32 tick = SDL_GetTicks();

	if(global.showfps == 0) return;


	sprintf(buf, "%s ... %d fps.\n", 
		SPS_State_ToString(global.state),
		1000 * global.frames / (tick - global.startticks));
	fprintf(stderr, buf);


	if(tick - global.startticks < 1000) return;
	font = GetFont("1");
	if(font == NULL) return;


	SDL_UnMapRGB(global.display->format, &c, context->forecolor);
	surface = TTF_RenderUTF8_Blended(font, buf, c);

	r.x = global.display->w - surface->w;
	r.y = global.display->h - surface->h;
	r.w = surface->w;
	r.h = surface->h;
	
	SDL_FillRect(global.display, &r, context->backcolor);
	SDL_BlitSurface(surface, NULL, global.display, &r);

	global.startticks = tick;
	global.frames = 0;
}

/* ---------------------------------------- */

void SPS_Fill()
{
	fprintf(stderr, "clear\n");
	SDL_FillRect(context->surface, NULL, context->backcolor);
}

void SPS_Pause(int timer)
{
	int infinite = 0;
	fprintf(stderr, "pause\n");

	if(timer == -1) infinite = 1;
	SPS_SetState(SPS_STATE_PAUSE);

	while(1){
		if(timer <= 0 && infinite == 0) break;
		if(global.state != SPS_STATE_PAUSE) break;
		SDL_Delay(20);
		timer -= 20;
	}
}

void SPS_MoveCursor(int x, int y)
{
	fprintf(stderr, "movecursor(%d,%d)\n", x, y);
	context->cursor.x = RatioToPixelH(x);
	context->cursor.y = RatioToPixelV(y);
}

void SPS_PutImage(const char* path)
{
	
}

static void do_render(SPS_Chunk_Context* chunk)
{
	SDL_Surface* rendered;
	gchar* utftext = NULL;
	if(chunk->ucs4text == NULL){
		rendered = chunk->surface;
	} else {
		TTF_SetFontStyle(chunk->font, chunk->style);

		utftext = g_ucs4_to_utf8(chunk->ucs4text, -1, NULL, NULL, NULL);
		rendered = TTF_RenderUTF8_Blended(chunk->font, utftext, chunk->color);
	}

	if(rendered != NULL){
		SDL_BlitSurface(rendered, NULL, context->surface, &context->cursor);
		context->cursor.x += rendered->w;
	} else {
		fprintf(stderr, "TTF_RenderUTF8_Blend: %s\n", SDL_GetError());
	}
	if(utftext == NULL) g_free(utftext);
}

void SPS_PutText(const char* text, const char* fontindex, SPS_VERTICAL_ALIGN align)
{
	SDL_Color c;
	TTF_Font* font;

	font = GetFont(fontindex);
	if(font == NULL) {
		fprintf(stderr, "GetFont == NULL\n");
		return;
	}
	SDL_UnMapRGB(context->surface->format, &c, context->forecolor);

	{
		size_t source_len, destination_len;
		char* tmp_input, * tmp_output;
		const char* ptmp_input;
		char* ptmp_output;
		int result;
		SDL_Surface* rendered = NULL;

		source_len = strlen(text) + 1;
		destination_len = source_len * 2;

		tmp_input = (char*)strdup(text);
		if(tmp_input == NULL) return;
		strncpy(tmp_input, text, source_len - 1);
		tmp_input[source_len - 1] = '\0';
		tmp_output = (char*)malloc(destination_len);
		if(tmp_output == NULL){
			free(tmp_input);
			return;
		}
		ptmp_input = tmp_input;
		ptmp_output = tmp_output;
		result = iconv(context->iconv_context, &ptmp_input,
			       &source_len, &ptmp_output, &destination_len);
		if(result == -1){
			fprintf(stderr, "iconv: error\n");
		} else {
			SPS_Chunk_Context* chunk;
			chunk = g_malloc(sizeof(SPS_Chunk_Context));
			if(chunk != NULL){
				chunk->ucs4text = g_utf8_to_ucs4(tmp_output, -1, NULL, NULL, NULL);
				chunk->font = font;
				chunk->color = c;
				chunk->style = TTF_GetFontStyle(font);
				chunk->done = 0;
				chunk->valign = align;
				chunk->surface = NULL;
			}

			if(context->in_linerender == 0){
				do_render(chunk);
				g_free(chunk);
			} else {
				context->line = g_list_append(context->line, chunk);
			}
		}
		free(tmp_input);
		free(tmp_output);
	}
	return;
}

void SPS_ForeColor(const char* color)
{
	context->forecolor = ParseColor(color);
}

void SPS_BackColor(const char* color)
{
	context->backcolor = ParseColor(color);
}

void SPS_PutHBar(double thick)
{
	int height;
	SDL_Rect r;

	height = RatioToPixelV(thick);
	r.x = context->hmargin;
	r.y = context->cursor.y;
	r.w = context->surface->w - r.x;
	r.h = height;
	fprintf(stderr, "puthbar: %f -> (%d, %d - %d x %d)\n", thick, r.x, r.y, r.w, r.h);
	SDL_FillRect(context->surface, &r, context->forecolor);

	context->cursor.x = 0;
	context->cursor.y += height;
}

void SPS_MoveTo(int x, int y)
{
	int width, height;
	width = RatioToPixelH(x);
	height = RatioToPixelV(y);
	fprintf(stderr, "moveto: %d, %d -> %d, %d\n", x, y, width, height);
	context->cursor.x += width;
	context->cursor.y += height;
}

int GetTopLeftY(SPS_VERTICAL_ALIGN valign, int h, int maxheight)
{
	switch(valign){
	case SPS_VERTICAL_ALIGN_TOP:
		return 0;
	case SPS_VERTICAL_ALIGN_CENTER:
		return (maxheight - h) / 2;
	case SPS_VERTICAL_ALIGN_BOTTOM:
		return maxheight - h;
	}
	return 0;
}

int GetTopLeftX(SPS_HORIZONTAL_ALIGN halign, int w, int maxwidth)
{
	switch(halign)
	{
	case SPS_HORIZONTAL_ALIGN_LEFT:
		return 0;
	case SPS_HORIZONTAL_ALIGN_CENTER:
		return (maxwidth - w) / 2;
	case SPS_HORIZONTAL_ALIGN_RIGHT:
		return maxwidth - w;
	}
	return 0;
}
int TTF_WidthUCS4(TTF_Font* font, gunichar* ucs4text, int width)
{
	int w, h;
	int len;
	int score;
	gchar* utf8text;
	gunichar* u;


	len = gunistr_len(ucs4text);
	u = gunistr_substring(ucs4text, 0, len);

	while(1){
		utf8text = g_ucs4_to_utf8(u, -1, NULL, NULL, NULL);
		TTF_SizeUTF8(font, utf8text, &w, &h);
		g_free(utf8text);

		score = width - w;
		if(score >= 0){
			/* OK */
			break;
		} else {
			gunichar* oldu = u;
			int oldlen = len;
			len += score / TTF_FontHeight(font) - 1;
			fprintf(stderr, "shorten: %d->%d\n", oldlen, len);

			if(len <= 0){
				break;
			} else {
				u = gunistr_substring(oldu, 0, len);
				g_free(oldu);
			}
		}
	}

	return len;

}

SDL_Surface* TTF_RenderUCS4_Blended(TTF_Font* font, gunichar* ucs4text, SDL_Color c)
{
	gchar* utf8text;
	SDL_Surface* result;

	utf8text = g_ucs4_to_utf8(ucs4text, -1, NULL, NULL, NULL);
	result = TTF_RenderUTF8_Blended(font, utf8text, c);
	g_free(utf8text);
	return result;
}

/* static */ typedef struct {
	SDL_Surface* surface;
	SPS_VERTICAL_ALIGN valign;
} ichunk;


static SDL_Surface* do_render_now(GList* surfaces, int w, int h)
{
	SDL_Surface* dst;
	SDL_Rect r;
	GList* p;

	dst = SDL_CreateRGBSurface(context->surface->flags,
				   w,
				   h,
				   context->surface->format->BitsPerPixel,
				   context->surface->format->Rmask,
				   context->surface->format->Gmask,
				   context->surface->format->Bmask,
				   context->surface->format->Amask);
	{
		SDL_Rect rr;
		rr.x = context->cursor.x;
		rr.y = context->cursor.y;
		rr.w = w;
		rr.h = h;

		SDL_BlitSurface(context->surface, &rr, dst, NULL);
	}

	r.x = 0;

	if(dst != NULL){				   
		G_LIST_FOREACH(surfaces, p){
			ichunk* chunk = (ichunk*)p->data;
			r.y = GetTopLeftY(chunk->valign, chunk->surface->h, h);

			SDL_BlitSurface(chunk->surface, NULL, dst, &r);
			r.x += chunk->surface->w;
			SDL_FreeSurface(chunk->surface);
			g_free(chunk);
		}
	}
	return dst;
}

static void render_now(GList* surfaces, int totalw, int maxh)
{
	SDL_Rect r;
	SDL_Surface* surface = 
		do_render_now(surfaces, totalw, maxh);
	
	r.x = context->cursor.x + GetTopLeftX(context->halign, 
			  totalw,
			  context->surface->w);
	r.y = context->cursor.y;
	SDL_BlitSurface(surface, NULL, context->surface, &r);
	SDL_FreeSurface(surface);
	
	context->cursor.x = 0;
	context->cursor.y += maxh;
}

void SPS_EndLine()
{
	GList* p;
	GList* surfaces = NULL;

	int x, y, w, h, totalw, maxh;
	int maxw;


	x = context->cursor.x;
	y = context->cursor.y;
	maxw = context->surface->w;

	totalw = maxh = 0;

 retry_all:

	G_LIST_FOREACH(context->line, p){
		ichunk* ic;

		SPS_Chunk_Context* chunk = p->data;
		int ww, l, len;
		SDL_Surface* surface;
		gunichar* t;
		if(chunk->done) continue;

		len = gunistr_len(chunk->ucs4text);
		if(len == 0) continue;
	retry:
		w = maxw - x;
		l = TTF_WidthUCS4(chunk->font, chunk->ucs4text, w);
		fprintf(stderr, "%d(/%d), %d-> %d\n", w, TTF_FontHeight(chunk->font), len, l);
		if(l <= 0){
			/* new line */
			render_now(surfaces, totalw, maxh);
			surfaces = NULL;

			y += maxh;
			x = 0;
			totalw = maxh = 0;

			goto retry;
		}

		chunk->done = 1;

		t = gunistr_substring(chunk->ucs4text, 0, l);
		TTF_SetFontStyle(chunk->font, chunk->style);
		surface = TTF_RenderUCS4_Blended(chunk->font, 
						 t,
						 chunk->color);

		g_free(t);

		x += surface->w;
		totalw += surface->w;
		if(maxh < surface->h) maxh = surface->h;

		ic = g_malloc(sizeof(ichunk));
		ic->surface = surface;
		ic->valign = chunk->valign;
		surfaces = g_list_append(surfaces, ic);
				
		if(l < len){
			/* fold */
			SPS_Chunk_Context* newchunk;
			render_now(surfaces, totalw, maxh);
			surfaces = NULL;
			x = 0;
			y += maxh;

			totalw = maxh = 0;

			newchunk = g_malloc(sizeof(SPS_Chunk_Context));

			newchunk->font = chunk->font;
			newchunk->color = chunk->color;
			newchunk->style = chunk->style;
			newchunk->ucs4text = gunistr_substring(chunk->ucs4text, l, len - l);
			newchunk->valign = chunk->valign;
			newchunk->done = 0;
			newchunk->surface = NULL;
			context->line = g_list_prepend(context->line, newchunk);
			goto retry_all;
		}

	}
	if(g_list_first(surfaces)){
		render_now(surfaces, totalw, maxh);
		surfaces = NULL;
		x = 0;
		y += maxh;
	}
}

void SPS_BeginLine(const char* halign)
{
	context->in_linerender = 1;
	context->halign = SPS_Horizontal_Align_Parse(halign);
	GList* p;
	fprintf(stderr, "beginline\n");
	context->cursor.x = context->hmargin;

	if(context->line != NULL){
		G_LIST_FOREACH(context->line, p){

		}
		g_list_free(context->line);
	}
	context->line = NULL;
}


void SPS_FontStyle(const char* fontindex, const char* expression)
{
	int style;
	TTF_Font* font = GetFont(fontindex);
	if(font == NULL) return;
	style = TTF_Style_Parse(expression);
	TTF_SetFontStyle(font, style);
}

void SPS_HMargin(int m)
{
	context->hmargin = RatioToPixelH(m);
}

void SPS_DefIcon(const char* no, const char* filename)
{
	gpointer oldvalue, oldkey;
	SDL_Surface* surface;
	surface = IMG_Load(filename);
	if(surface == NULL) return;
	if(g_hash_table_lookup_extended(context->icons, no, &oldkey, &oldvalue)){
		g_hash_table_remove(context->icons, no);
		g_free(oldkey);
		SDL_FreeSurface((SDL_Surface*)oldvalue);
	}
	g_hash_table_insert(context->icons, g_strdup(no), surface);
	fprintf(stderr, "icon(%s) = %s\n", no, filename);
}

void SPS_PutIcon(const char* iconindex, SPS_VERTICAL_ALIGN align)
{
	SDL_Surface* surface;
	SPS_Chunk_Context* chunk;

	surface = GetIcon(iconindex);
	if(surface == NULL) return;
	fprintf(stderr, "puticon(%s)\n", iconindex);

	chunk = g_malloc(sizeof(SPS_Chunk_Context));
	if(chunk != NULL){
		chunk->ucs4text = NULL;
		chunk->surface = surface;
		context->line = g_list_append(context->line, chunk);
	}
	return;
}


void SPS_BeginPage()
{
	fprintf(stderr, "beingpage()\n");
	context->cursor.x = 0;
	context->cursor.y = 0;
	context->hmargin = 0;
}
